import { Ordering } from "./Ordering";
import { FromItem, Row, HardRow } from "./FromFactor";
import { RetrievalQuery } from "./Queries/RetrievalQuery";
import { Table } from "./Table";
import { AnyType, BooleanType, GetInType, IntegerType, Json, Record, RecordTypeToJson, TextType, VoidType, Type, StandaloneNullType, GetOutType } from './Types';
export declare type ExpressionTypeOf<T extends Expression<any>> = T["type"];
export declare abstract class Expression<T extends AnyType> {
    readonly type: T;
    constructor(type: T);
    get precedenceLevel(): number;
    cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
    narrow<T2 extends GetOutType<T>>(): Expression<Type<GetInType<T>, T2, T["_brand"]>>;
    as<TName extends string>(name: TName): NamedExpression<TName, T>;
    isEqualTo(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isNotEqualTo(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isLessThan(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isAtMost(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isGreaterThan(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isAtLeast(other: ExpressionOrInputValue<T>): Expression<BooleanType>;
    isNull(): Expression<BooleanType>;
    isNotNull(): Expression<BooleanType>;
    and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
    or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
    not(this: Expression<BooleanType>): Expression<BooleanType>;
    isIn(values: ExpressionOrInputValue<T>[]): Expression<BooleanType>;
    isInQuery<TRow extends {
        [name: string]: AnyType;
    }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
    abs(this: Expression<IntegerType>): Expression<IntegerType>;
    minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
    plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
    mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
    div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
    toUpper(this: Expression<TextType>): Expression<TextType>;
    toLower(this: Expression<TextType>): Expression<TextType>;
    sum(this: Expression<IntegerType>): Expression<IntegerType>;
    count(this: Expression<any>): Expression<IntegerType>;
    min(this: Expression<IntegerType>): Expression<IntegerType>;
    max(this: Expression<IntegerType>): Expression<IntegerType>;
    distinct(): Expression<T>;
    isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
    asc(): Ordering<Expression<T>>;
    desc(): Ordering<Expression<T>>;
    toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<RecordTypeToJson<Record<TColumns>>>>;
    prop<TKey extends keyof GetInType<ExpressionTypeOf<this>>>(this: Expression<Json<any>>, key: TKey): Expression<Json<GetInType<ExpressionTypeOf<this>>[TKey]>>;
}
export declare class CastExpression<T extends AnyType> extends Expression<T> {
    readonly expression: Expression<AnyType>;
    readonly isHiddenCast: boolean;
    get precedenceLevel(): number;
    constructor(expression: Expression<AnyType>, newType: T, isHiddenCast: boolean);
}
export declare class JsonPropertyAccess extends Expression<Json<any>> {
    readonly expression: Expression<Json<any>>;
    readonly key: string;
    constructor(expression: Expression<Json<any>>, key: string);
}
export declare class FromItemExpression<T extends Row> extends Expression<Record<T>> {
    readonly fromItem: FromItem<T>;
    constructor(fromItem: FromItem<T>);
}
export declare function normalize<T extends AnyType>(type: T, expr: ExpressionOrInputValue<T>): Expression<T>;
export declare function not(expression: Expression<BooleanType>): Expression<BooleanType>;
export declare function and(expression1: Expression<BooleanType>, ...expressions: Expression<BooleanType>[]): Expression<BooleanType>;
export declare function and(...expressions: (Expression<BooleanType> | undefined)[]): Expression<BooleanType> | undefined;
export declare function or(expression1: Expression<BooleanType>, ...expressions: Expression<BooleanType>[]): Expression<BooleanType>;
export declare function or(...expressions: (Expression<BooleanType> | undefined)[]): Expression<BooleanType> | undefined;
export declare function nullif<T extends AnyType>(expression1: Expression<T>, expression2: ExpressionOrInputValue<T>): Expression<T>;
export declare function concat(...expressions: ExpressionOrInputValue<TextType>[]): Expression<TextType>;
export declare function coalesce<T extends AnyType>(...expressions: Expression<T>[]): Expression<T>;
export declare function val(value: null, preferEscaping?: boolean): ValueExpression<StandaloneNullType>;
export declare function val(value: string, preferEscaping?: boolean): ValueExpression<TextType>;
export declare function val(value: boolean, preferEscaping?: boolean): ValueExpression<BooleanType>;
export declare function val(value: number, preferEscaping?: boolean): ValueExpression<IntegerType>;
export declare function val<T extends AnyType>(value: GetInType<T>, type: T, preferEscaping?: boolean): ValueExpression<T>;
export declare function jsonVal<T>(value: T): ValueExpression<Json<T>>;
export declare function defaultValue(): DefaultExpression;
export declare function toCondition<TColumns extends HardRow>(fromItem: FromItem<TColumns> | undefined, args: (Expression<BooleanType>[]) | [Partial<MapExpressionOrInputValue<TColumns>>]): Expression<BooleanType> | undefined;
export declare type NamedExpressionNameOf<T extends NamedExpression<any, any>> = T["name"];
export declare class NamedExpression<TColumnName extends string, T extends AnyType> extends Expression<T> {
    readonly name: TColumnName;
    constructor(name: TColumnName, type: T);
}
export declare class Column<TColumnName extends string, T extends AnyType> extends NamedExpression<TColumnName, T> {
    private _fromItem;
    get fromItem(): FromItem<any>;
    constructor(name: TColumnName, type: T, fromItemSetterProvider: (setter: (fromItem: FromItem<any>) => void) => void);
}
export declare class TableColumn<TColumnName extends string, T extends AnyType> extends Column<TColumnName, T> {
    constructor(name: TColumnName, type: T, fromItemSetterProvider: (setter: (fromItem: Table<any, any>) => void) => void);
}
export declare class AsColumn<TColumnName extends string, T extends AnyType> extends Column<TColumnName, T> {
    constructor(name: TColumnName, type: T, fromItemSetterProvider: (setter: (fromItem: Table<any, any>) => void) => void);
}
export declare class ColumnBoundToExpression<TColumnName extends string, T extends AnyType> extends Column<TColumnName, T> {
    private readonly expression;
    constructor(expression: NamedExpression<TColumnName, T>, fromItemSetterProvider: (setter: (fromItem: FromItem<any>) => void) => void);
}
export declare class AllExpression<TColumns extends Row> extends Expression<VoidType> {
    readonly fromItem: FromItem<TColumns>;
    constructor(fromItem: FromItem<TColumns>);
}
export declare class NamedExpressionWrapper<TColumnName extends string, T extends AnyType> extends NamedExpression<TColumnName, T> {
    readonly expression: Expression<T>;
    constructor(name: TColumnName, expression: Expression<T>);
}
export declare class RetrievalQueryAsExpression<T extends AnyType> extends Expression<T> {
    readonly query: RetrievalQuery<any, any>;
    constructor(query: RetrievalQuery<any, any>, type: T);
}
export declare class ValueExpression<T extends AnyType> extends Expression<T> {
    readonly value: GetInType<T>;
    readonly preferEscaping: boolean;
    constructor(type: T, value: GetInType<T>, preferEscaping?: boolean);
}
export declare type ExpressionOrInputValue<T extends AnyType> = GetInType<T> | Expression<T>;
export declare type MapExpressionOrInputValue<T extends {
    [key: string]: any;
}> = {
    [TKey in keyof T]: ExpressionOrInputValue<T[TKey]>;
};
export declare class Variable<T extends AnyType> extends Expression<T> {
    readonly name: string;
    constructor(name: string, type: T);
}
export declare abstract class BinaryOperatorExpression<TLeft extends AnyType, TRight extends AnyType, TResult extends AnyType> extends Expression<TResult> {
    readonly left: Expression<TLeft>;
    readonly right: Expression<TRight>;
    constructor(left: Expression<TLeft>, right: Expression<TRight>, type: TResult);
    abstract get operator(): string;
}
export interface ConcreteBinaryExpression<TLeft extends AnyType, TRight extends AnyType, TResult extends AnyType> extends BinaryOperatorExpression<TLeft, TRight, TResult> {
    operator: string;
    precedenceLevel: number;
}
export interface ConcreteBinaryExpressionStatic<TLeft extends AnyType, TRight extends AnyType, TResult extends AnyType> {
    new (...args: any[]): ConcreteBinaryExpression<TLeft, TRight, TResult>;
}
export declare function ConcreteBinaryExpression<TResult extends AnyType>(symbol: string, resultType: TResult, precedenceLevel: number): {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: TResult;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends GetOutType<TResult>>(): Expression<Type<GetInType<TResult>, T2_1, TResult["_brand"]>>;
        as<TName extends string>(name: TName): NamedExpression<TName, TResult>;
        isEqualTo(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<TResult>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<TResult>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<TResult>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<TResult>>;
        desc(): Ordering<Expression<TResult>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends keyof GetInType<TResult>>(this: Expression<Json<any>>, key: TKey): Expression<Json<GetInType<TResult>[TKey]>>;
    };
};
declare const AdditionExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: IntegerType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends number>(): Expression<Type<number, T2_1, "integer">>;
        as<TName extends string>(name: TName): NamedExpression<TName, IntegerType>;
        isEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<IntegerType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<IntegerType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<IntegerType>>;
        desc(): Ordering<Expression<IntegerType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "toString" | "valueOf" | "toFixed" | "toExponential" | "toPrecision" | "toLocaleString">(this: Expression<Json<any>>, key: TKey): Expression<Json<number[TKey]>>;
    };
};
export declare class AdditionExpression extends AdditionExpression_base<IntegerType, IntegerType> {
}
declare const SubtractionExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: IntegerType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends number>(): Expression<Type<number, T2_1, "integer">>;
        as<TName extends string>(name: TName): NamedExpression<TName, IntegerType>;
        isEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<IntegerType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<IntegerType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<IntegerType>>;
        desc(): Ordering<Expression<IntegerType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "toString" | "valueOf" | "toFixed" | "toExponential" | "toPrecision" | "toLocaleString">(this: Expression<Json<any>>, key: TKey): Expression<Json<number[TKey]>>;
    };
};
export declare class SubtractionExpression extends SubtractionExpression_base<IntegerType, IntegerType> {
}
declare const MultiplicationExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: IntegerType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends number>(): Expression<Type<number, T2_1, "integer">>;
        as<TName extends string>(name: TName): NamedExpression<TName, IntegerType>;
        isEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<IntegerType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<IntegerType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<IntegerType>>;
        desc(): Ordering<Expression<IntegerType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "toString" | "valueOf" | "toFixed" | "toExponential" | "toPrecision" | "toLocaleString">(this: Expression<Json<any>>, key: TKey): Expression<Json<number[TKey]>>;
    };
};
export declare class MultiplicationExpression extends MultiplicationExpression_base<IntegerType, IntegerType> {
}
declare const DivisionExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: IntegerType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends number>(): Expression<Type<number, T2_1, "integer">>;
        as<TName extends string>(name: TName): NamedExpression<TName, IntegerType>;
        isEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<IntegerType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<IntegerType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<IntegerType>>;
        desc(): Ordering<Expression<IntegerType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "toString" | "valueOf" | "toFixed" | "toExponential" | "toPrecision" | "toLocaleString">(this: Expression<Json<any>>, key: TKey): Expression<Json<number[TKey]>>;
    };
};
export declare class DivisionExpression extends DivisionExpression_base<IntegerType, IntegerType> {
}
declare const ModulusExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: IntegerType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends number>(): Expression<Type<number, T2_1, "integer">>;
        as<TName extends string>(name: TName): NamedExpression<TName, IntegerType>;
        isEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<IntegerType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<IntegerType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<IntegerType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<IntegerType>>;
        desc(): Ordering<Expression<IntegerType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "toString" | "valueOf" | "toFixed" | "toExponential" | "toPrecision" | "toLocaleString">(this: Expression<Json<any>>, key: TKey): Expression<Json<number[TKey]>>;
    };
};
export declare class ModulusExpression extends ModulusExpression_base<IntegerType, IntegerType> {
}
declare const EqualsExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class EqualsExpression<T extends AnyType> extends EqualsExpression_base<T, T> {
}
declare const UnequalsExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class UnequalsExpression<T extends AnyType> extends UnequalsExpression_base<T, T> {
}
declare const GreaterExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class GreaterExpression<T extends AnyType> extends GreaterExpression_base<T, T> {
}
declare const LessExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class LessExpression<T extends AnyType> extends LessExpression_base<T, T> {
}
declare const GreaterOrEqualExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class GreaterOrEqualExpression<T extends AnyType> extends GreaterOrEqualExpression_base<T, T> {
}
declare const LessOrEqualExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class LessOrEqualExpression<T extends AnyType> extends LessOrEqualExpression_base<T, T> {
}
declare const NotLessExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class NotLessExpression<T extends AnyType> extends NotLessExpression_base<T, T> {
}
declare const NotGreaterExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class NotGreaterExpression<T extends AnyType> extends NotGreaterExpression_base<T, T> {
}
declare const OrExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class OrExpression extends OrExpression_base<BooleanType, BooleanType> {
}
declare const AndExpression_base: {
    new <TLeft extends AnyType, TRight extends AnyType>(left: Expression<TLeft>, right: Expression<TRight>): {
        readonly operator: string;
        readonly precedenceLevel: number;
        readonly left: Expression<TLeft>;
        readonly right: Expression<TRight>;
        readonly type: BooleanType;
        cast<T2 extends AnyType>(type: T2, isCastHidden?: boolean): Expression<T2>;
        narrow<T2_1 extends boolean>(): Expression<Type<boolean, T2_1, "boolean">>;
        as<TName extends string>(name: TName): NamedExpression<TName, BooleanType>;
        isEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNotEqualTo(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isLessThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtMost(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isGreaterThan(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isAtLeast(other: ExpressionOrInputValue<BooleanType>): Expression<BooleanType>;
        isNull(): Expression<BooleanType>;
        isNotNull(): Expression<BooleanType>;
        and(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        or(this: Expression<BooleanType>, other: Expression<BooleanType>): Expression<BooleanType>;
        not(this: Expression<BooleanType>): Expression<BooleanType>;
        isIn(values: ExpressionOrInputValue<BooleanType>[]): Expression<BooleanType>;
        isInQuery<TRow extends {
            [name: string]: AnyType;
        }, TSingleColumn extends keyof TRow>(this: Expression<TRow[TSingleColumn]>, values: RetrievalQuery<TRow, TSingleColumn>): Expression<BooleanType>;
        abs(this: Expression<IntegerType>): Expression<IntegerType>;
        minus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        plus(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        mult(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        div(this: Expression<IntegerType>, other: ExpressionOrInputValue<IntegerType>): Expression<IntegerType>;
        toUpper(this: Expression<TextType>): Expression<TextType>;
        toLower(this: Expression<TextType>): Expression<TextType>;
        sum(this: Expression<IntegerType>): Expression<IntegerType>;
        count(this: Expression<any>): Expression<IntegerType>;
        min(this: Expression<IntegerType>): Expression<IntegerType>;
        max(this: Expression<IntegerType>): Expression<IntegerType>;
        distinct(): Expression<BooleanType>;
        isLike(this: Expression<TextType>, other: ExpressionOrInputValue<TextType>): Expression<BooleanType>;
        asc(): Ordering<Expression<BooleanType>>;
        desc(): Ordering<Expression<BooleanType>>;
        toJson<TColumns>(this: Expression<Record<TColumns>>): Expression<Json<{ [TName_1 in keyof TColumns]: RecordTypeToJson<TColumns[TName_1]>; }>>;
        prop<TKey extends "valueOf">(this: Expression<Json<any>>, key: TKey): Expression<Json<boolean[TKey]>>;
    };
};
export declare class AndExpression extends AndExpression_base<BooleanType, BooleanType> {
}
export declare class LikeExpression extends Expression<BooleanType> {
    readonly argument: Expression<TextType>;
    readonly like: Expression<TextType>;
    get precedenceLevel(): number;
    constructor(argument: Expression<TextType>, like: Expression<TextType>);
}
export declare class IsInValuesExpression<T extends AnyType> extends Expression<BooleanType> {
    readonly argument: Expression<T>;
    readonly values: Expression<T>[];
    get precedenceLevel(): number;
    constructor(argument: Expression<T>, values: Expression<T>[]);
}
export declare class IsInQueryExpression<T extends AnyType> extends Expression<BooleanType> {
    readonly argument: Expression<T>;
    readonly query: RetrievalQuery<any, any>;
    get precedenceLevel(): number;
    constructor(argument: Expression<T>, query: RetrievalQuery<any, any>);
}
export declare class IsNullExpression extends Expression<BooleanType> {
    readonly argument: Expression<any>;
    constructor(argument: Expression<any>);
}
export declare class IsNotNullExpression extends Expression<BooleanType> {
    readonly argument: Expression<any>;
    constructor(argument: Expression<any>);
}
export declare class NotExpression extends Expression<BooleanType> {
    readonly argument: Expression<BooleanType>;
    get precedenceLevel(): number;
    constructor(argument: Expression<BooleanType>);
}
export declare class KnownFunctionInvocation<TResultType extends AnyType> extends Expression<TResultType> {
    readonly functionName: string;
    readonly args: Expression<any>[];
    constructor(functionName: string, args: Expression<any>[], resultType: TResultType);
}
export declare class DefaultExpression extends Expression<AnyType> {
    constructor();
}
export {};
