Object.defineProperty(exports, "__esModule", { value: true });
exports.DefaultExpression = exports.KnownFunctionInvocation = exports.NotExpression = exports.IsNotNullExpression = exports.IsNullExpression = exports.IsInQueryExpression = exports.IsInValuesExpression = exports.LikeExpression = exports.AndExpression = exports.OrExpression = exports.NotGreaterExpression = exports.NotLessExpression = exports.LessOrEqualExpression = exports.GreaterOrEqualExpression = exports.LessExpression = exports.GreaterExpression = exports.UnequalsExpression = exports.EqualsExpression = exports.ModulusExpression = exports.DivisionExpression = exports.MultiplicationExpression = exports.SubtractionExpression = exports.AdditionExpression = exports.ConcreteBinaryExpression = exports.BinaryOperatorExpression = exports.Variable = exports.ValueExpression = exports.RetrievalQueryAsExpression = exports.NamedExpressionWrapper = exports.AllExpression = exports.ColumnBoundToExpression = exports.AsColumn = exports.TableColumn = exports.Column = exports.NamedExpression = exports.toCondition = exports.defaultValue = exports.jsonVal = exports.val = exports.coalesce = exports.concat = exports.nullif = exports.or = exports.and = exports.not = exports.normalize = exports.FromItemExpression = exports.JsonPropertyAccess = exports.CastExpression = exports.Expression = void 0;
const FromFactor_1 = require("./FromFactor");
const Types_1 = require("./Types");
class Expression {
    constructor(type) {
        if (!type || !type.serialize)
            throw new Error(`No type given: ${type}.`);
        this.type = type;
    }
    get precedenceLevel() { return 0; }
    cast(type, isCastHidden = false) { return new CastExpression(this, type, isCastHidden); }
    narrow() { return this; }
    as(name) {
        return new NamedExpressionWrapper(name, this);
    }
    isEqualTo(other) {
        return new EqualsExpression(this, normalize(this.type, other));
    }
    isNotEqualTo(other) {
        return new UnequalsExpression(this, normalize(this.type, other));
    }
    isLessThan(other) {
        return new LessExpression(this, normalize(this.type, other));
    }
    isAtMost(other) {
        return new LessOrEqualExpression(this, normalize(this.type, other));
    }
    isGreaterThan(other) {
        return new GreaterExpression(this, normalize(this.type, other));
    }
    isAtLeast(other) {
        return new GreaterOrEqualExpression(this, normalize(this.type, other));
    }
    isNull() { return new IsNullExpression(this); }
    isNotNull() { return new IsNotNullExpression(this); }
    and(other) {
        return new AndExpression(this, other);
    }
    or(other) {
        return new OrExpression(this, other);
    }
    not() { return new NotExpression(this); }
    isIn(values) { return new IsInValuesExpression(this, values.map(v => normalize(this.type, v))); }
    isInQuery(values) { return new IsInQueryExpression(this, values); }
    abs() { return new KnownFunctionInvocation("abs", [this], Types_1.tInteger); }
    minus(other) {
        return new SubtractionExpression(this, normalize(Types_1.tInteger, other));
    }
    plus(other) {
        return new AdditionExpression(this, normalize(Types_1.tInteger, other));
    }
    mult(other) {
        return new MultiplicationExpression(this, normalize(Types_1.tInteger, other));
    }
    div(other) {
        return new DivisionExpression(this, normalize(Types_1.tInteger, other));
    }
    toUpper() { return new KnownFunctionInvocation("upper", [this], Types_1.tText); }
    toLower() { return new KnownFunctionInvocation("lower", [this], Types_1.tText); }
    /* Grouping functions */
    sum() { return new KnownFunctionInvocation("sum", [this], Types_1.tInteger); }
    count() { return new KnownFunctionInvocation("count", [this], Types_1.tInteger); }
    min() { return new KnownFunctionInvocation("min", [this], Types_1.tInteger); }
    max() { return new KnownFunctionInvocation("max", [this], Types_1.tInteger); }
    distinct() { return new KnownFunctionInvocation("distinct", [this], this.type); }
    isLike(other) { return new LikeExpression(this, normalize(this.type, other)); }
    asc() { return { asc: this }; }
    ;
    desc() { return { desc: this }; }
    ;
    toJson() {
        return new KnownFunctionInvocation("row_to_json", [this], (0, Types_1.tJson)());
    }
    prop(key) {
        return new JsonPropertyAccess(this, key);
    }
}
exports.Expression = Expression;
class CastExpression extends Expression {
    constructor(expression, newType, isHiddenCast) {
        super(newType);
        this.expression = expression;
        this.isHiddenCast = isHiddenCast;
    }
    get precedenceLevel() { return this.expression.precedenceLevel; }
}
exports.CastExpression = CastExpression;
class JsonPropertyAccess extends Expression {
    constructor(expression, key) {
        super((0, Types_1.tJson)());
        this.expression = expression;
        this.key = key;
    }
}
exports.JsonPropertyAccess = JsonPropertyAccess;
class FromItemExpression extends Expression {
    constructor(fromItem) {
        super((0, Types_1.tRecord)());
        this.fromItem = fromItem;
    }
}
exports.FromItemExpression = FromItemExpression;
function normalize(type, expr) {
    if (expr instanceof Expression)
        return expr;
    return new ValueExpression(type, expr);
}
exports.normalize = normalize;
function not(expression) {
    return new NotExpression(expression);
}
exports.not = not;
function and(...expressions) {
    return expressions
        .filter(expr => !!expr)
        .reduce((prev, cur) => prev ? prev.and(cur) : cur, undefined);
}
exports.and = and;
function or(...expressions) {
    return expressions
        .filter(expr => !!expr)
        .reduce((prev, cur) => prev ? prev.or(cur) : cur, undefined);
}
exports.or = or;
function nullif(expression1, expression2) {
    return new KnownFunctionInvocation("nullif", [expression1, normalize(expression1.type, expression2)], expression1.type);
}
exports.nullif = nullif;
function concat(...expressions) {
    return new KnownFunctionInvocation("concat", expressions.map(e => normalize(Types_1.tText, e)), Types_1.tText);
}
exports.concat = concat;
function coalesce(...expressions) {
    return new KnownFunctionInvocation("coalesce", expressions, expressions[expressions.length - 1].type);
}
exports.coalesce = coalesce;
function val(value, ...other) {
    let preferEscaping;
    if (other[0] instanceof Types_1.Type) {
        preferEscaping = !!other[1];
        return new ValueExpression(other[0], value, preferEscaping);
    }
    preferEscaping = !!other[0];
    let t;
    if (typeof value === "string")
        t = Types_1.tText;
    else if (typeof value === "boolean")
        t = Types_1.tBoolean;
    else if (typeof value === "number")
        t = Types_1.tInteger;
    else if (value === null)
        t = Types_1.tNull;
    else
        throw new Error(`Unsupported value: '${value}'.`);
    return new ValueExpression(t, value, preferEscaping);
}
exports.val = val;
function jsonVal(value) {
    return val(value, (0, Types_1.tJson)());
}
exports.jsonVal = jsonVal;
function defaultValue() { return new DefaultExpression(); }
exports.defaultValue = defaultValue;
function toCondition(fromItem, args) {
    const firstArg = args[0];
    if (firstArg && !(firstArg instanceof Expression)) {
        if (typeof firstArg !== "object" || Array.isArray(firstArg))
            throw new Error(`Where argument must be an object, but was '${firstArg}'.`);
        return and(...Object.keys(firstArg).map(columnName => {
            const expression = firstArg[columnName];
            if (!fromItem)
                throw new Error("No table to select from specified.");
            const column = (0, FromFactor_1.getColumn)(fromItem, columnName);
            return column.isEqualTo(expression);
        }));
    }
    else {
        return and(...args.map((arg, argIdx) => {
            if (!(arg instanceof Expression))
                throw new Error(`Argument at position '${argIdx}' is not of type expression.`);
            return arg;
        }));
    }
}
exports.toCondition = toCondition;
class NamedExpression extends Expression {
    constructor(name, type) {
        super(type);
        this.name = name;
    }
}
exports.NamedExpression = NamedExpression;
class Column extends NamedExpression {
    constructor(name, type, fromItemSetterProvider) {
        super(name, type);
        fromItemSetterProvider(fromItem => this._fromItem = fromItem);
    }
    get fromItem() { return this._fromItem; }
}
exports.Column = Column;
class TableColumn extends Column {
    constructor(name, type, fromItemSetterProvider) {
        super(name, type, fromItemSetterProvider);
    }
}
exports.TableColumn = TableColumn;
class AsColumn extends Column {
    constructor(name, type, fromItemSetterProvider) {
        super(name, type, fromItemSetterProvider);
    }
}
exports.AsColumn = AsColumn;
class ColumnBoundToExpression extends Column {
    constructor(expression, fromItemSetterProvider) {
        super(expression.name, expression.type, fromItemSetterProvider);
        this.expression = expression;
    }
}
exports.ColumnBoundToExpression = ColumnBoundToExpression;
class AllExpression extends Expression {
    constructor(fromItem) {
        super(Types_1.tVoid);
        this.fromItem = fromItem;
    }
}
exports.AllExpression = AllExpression;
class NamedExpressionWrapper extends NamedExpression {
    constructor(name, expression) {
        super(name, expression.type);
        this.expression = expression;
    }
}
exports.NamedExpressionWrapper = NamedExpressionWrapper;
class RetrievalQueryAsExpression extends Expression {
    constructor(query, type) {
        super(type);
        this.query = query;
    }
}
exports.RetrievalQueryAsExpression = RetrievalQueryAsExpression;
class ValueExpression extends Expression {
    constructor(type, value, preferEscaping = false) {
        super(type);
        this.value = value;
        this.preferEscaping = preferEscaping;
    }
}
exports.ValueExpression = ValueExpression;
class Variable extends Expression {
    constructor(name, type) {
        super(type);
        this.name = name;
    }
}
exports.Variable = Variable;
class BinaryOperatorExpression extends Expression {
    constructor(left, right, type) {
        super(type);
        this.left = left;
        this.right = right;
    }
}
exports.BinaryOperatorExpression = BinaryOperatorExpression;
function ConcreteBinaryExpression(symbol, resultType, precedenceLevel) {
    return class ConcreteBinaryExpression extends BinaryOperatorExpression {
        constructor(left, right) { super(left, right, resultType); }
        get operator() { return symbol; }
        get precedenceLevel() { return precedenceLevel; }
    };
}
exports.ConcreteBinaryExpression = ConcreteBinaryExpression;
class AdditionExpression extends ConcreteBinaryExpression("+", Types_1.tInteger, 3) {
}
exports.AdditionExpression = AdditionExpression;
class SubtractionExpression extends ConcreteBinaryExpression("-", Types_1.tInteger, 3) {
}
exports.SubtractionExpression = SubtractionExpression;
class MultiplicationExpression extends ConcreteBinaryExpression("*", Types_1.tInteger, 2) {
}
exports.MultiplicationExpression = MultiplicationExpression;
class DivisionExpression extends ConcreteBinaryExpression("/", Types_1.tInteger, 2) {
}
exports.DivisionExpression = DivisionExpression;
class ModulusExpression extends ConcreteBinaryExpression("%", Types_1.tInteger, 2) {
}
exports.ModulusExpression = ModulusExpression;
class EqualsExpression extends ConcreteBinaryExpression("=", Types_1.tBoolean, 4) {
}
exports.EqualsExpression = EqualsExpression;
class UnequalsExpression extends ConcreteBinaryExpression("!=", Types_1.tBoolean, 4) {
}
exports.UnequalsExpression = UnequalsExpression;
class GreaterExpression extends ConcreteBinaryExpression(">", Types_1.tBoolean, 4) {
}
exports.GreaterExpression = GreaterExpression;
class LessExpression extends ConcreteBinaryExpression("<", Types_1.tBoolean, 4) {
}
exports.LessExpression = LessExpression;
class GreaterOrEqualExpression extends ConcreteBinaryExpression(">=", Types_1.tBoolean, 4) {
}
exports.GreaterOrEqualExpression = GreaterOrEqualExpression;
class LessOrEqualExpression extends ConcreteBinaryExpression("<=", Types_1.tBoolean, 4) {
}
exports.LessOrEqualExpression = LessOrEqualExpression;
class NotLessExpression extends ConcreteBinaryExpression("!<", Types_1.tBoolean, 4) {
}
exports.NotLessExpression = NotLessExpression;
class NotGreaterExpression extends ConcreteBinaryExpression("!>", Types_1.tBoolean, 4) {
}
exports.NotGreaterExpression = NotGreaterExpression;
class OrExpression extends ConcreteBinaryExpression("OR", Types_1.tBoolean, 7) {
}
exports.OrExpression = OrExpression;
class AndExpression extends ConcreteBinaryExpression("AND", Types_1.tBoolean, 6) {
}
exports.AndExpression = AndExpression;
class LikeExpression extends Expression {
    constructor(argument, like) {
        super(Types_1.tBoolean);
        this.argument = argument;
        this.like = like;
    }
    get precedenceLevel() { return 7; }
}
exports.LikeExpression = LikeExpression;
class IsInValuesExpression extends Expression {
    constructor(argument, values) {
        super(Types_1.tBoolean);
        this.argument = argument;
        this.values = values;
    }
    get precedenceLevel() { return 7; }
}
exports.IsInValuesExpression = IsInValuesExpression;
class IsInQueryExpression extends Expression {
    constructor(argument, query) {
        super(Types_1.tBoolean);
        this.argument = argument;
        this.query = query;
    }
    get precedenceLevel() { return 7; }
}
exports.IsInQueryExpression = IsInQueryExpression;
class IsNullExpression extends Expression {
    constructor(argument) {
        super(Types_1.tBoolean);
        this.argument = argument;
    }
}
exports.IsNullExpression = IsNullExpression;
class IsNotNullExpression extends Expression {
    constructor(argument) {
        super(Types_1.tBoolean);
        this.argument = argument;
    }
}
exports.IsNotNullExpression = IsNotNullExpression;
class NotExpression extends Expression {
    constructor(argument) {
        super(argument.type);
        this.argument = argument;
    }
    get precedenceLevel() { return 5; }
}
exports.NotExpression = NotExpression;
class KnownFunctionInvocation extends Expression {
    constructor(functionName, args, resultType) {
        super(resultType);
        this.functionName = functionName;
        this.args = args;
    }
}
exports.KnownFunctionInvocation = KnownFunctionInvocation;
class DefaultExpression extends Expression {
    constructor() { super(Types_1.tVoid); }
}
exports.DefaultExpression = DefaultExpression;
//# sourceMappingURL=Expressions.js.map