Object.defineProperty(exports, "__esModule", { value: true });
exports.DeleteQuery = exports.deleteFrom = void 0;
const FromFactor_1 = require("../FromFactor");
const Expressions_1 = require("../Expressions");
const Query_1 = require("./Query");
const Common_1 = require("./Common");
const JoinMixin_1 = require("./JoinMixin");
const Helpers_1 = require("../../Helpers");
/**
 * Creates a DELETE FROM statement.
 *
 * @param table The table to delete from.
 */
function deleteFrom(table) {
    return new DeleteQuery(table);
}
exports.deleteFrom = deleteFrom;
/**
 * Represents a DELETE FROM query.
 */
class DeleteQuery extends Query_1.Query {
    constructor(table) {
        super();
        this.table = table;
        // #endregion
        // #region call-macro joinMixin()
        this._from = undefined;
        this.lastFromItem = table;
    }
    getState() {
        return (0, Helpers_1.combine)(super.getState(), {
            table: this.table,
            using: this._from,
            whereCondition: this._whereCondition
        });
    }
    /**
     * Brings an additional query into scope that can be used to specify more complex delete conditions.
     *
     * @param fromItem The query to use.
     */
    using(fromItem) {
        this._from = FromFactor_1.FromFactor.crossJoin(this._from, fromItem);
        this.lastFromItem = fromItem;
        return this;
    }
    returning(...args) {
        (0, Common_1.handleSelect)(this.lastFromItem, args, this.returningColumns, this.selectedExpressions);
        return this;
    }
    where(...args) {
        const expression = (0, Expressions_1.toCondition)(this.lastFromItem, args);
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, expression);
        return this;
    }
    /**
    * Adds negated where conditions.
    */
    whereNot(condition, ...conditions) {
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, condition.not(), ...conditions.map(c => c.not()));
        return this;
    }
    /**
    * Performs a full join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * for (row j in joined): if (cur has no row that matches j)
    * 	yield row(null, j)
    * ```
    */
    fullJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorFullJoin, fromItem);
    }
    /**
    * Performs a left join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * ```
    */
    leftJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorLeftJoin, fromItem);
    }
    /**
    * Performs an inner join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * ```
    */
    innerJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorInnerJoin, fromItem);
    }
}
exports.DeleteQuery = DeleteQuery;
//# sourceMappingURL=DeleteQuery.js.map