Object.defineProperty(exports, "__esModule", { value: true });
exports.JoinConditionBuilder = exports.doJoin = void 0;
const Expressions_1 = require("../Expressions");
const FromFactor_1 = require("../FromFactor");
class JoinMixin {
    constructor() {
        // #region define-macro joinMixin()
        this._from = undefined;
        // #endregion
    }
    /**
    * Performs a full join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * for (row j in joined): if (cur has no row that matches j)
    * 	yield row(null, j)
    * ```
    */
    fullJoin(fromItem) {
        return doJoin(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorFullJoin, fromItem);
    }
    /**
    * Performs a left join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * ```
    */
    leftJoin(fromItem) {
        return doJoin(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorLeftJoin, fromItem);
    }
    /**
    * Performs an inner join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * ```
    */
    innerJoin(fromItem) {
        return doJoin(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorInnerJoin, fromItem);
    }
}
function doJoin(self, currentFrom, fromUpdater, ctor, fromItem) {
    return new JoinConditionBuilder(fromItem, (joined, condition) => {
        if (!currentFrom)
            throw new Error("A primary table must be selected before other tables can be joined.");
        fromUpdater(new ctor(currentFrom, joined, condition));
        return self;
    });
}
exports.doJoin = doJoin;
class JoinConditionBuilder {
    constructor(fromItem, onJoin) {
        this.fromItem = fromItem;
        this.onJoin = onJoin;
    }
    on(...conditions) {
        const condition = (0, Expressions_1.toCondition)(this.fromItem, conditions);
        if (!condition)
            throw new Error("No condition for joined was given.");
        return this.onJoin(this.fromItem, condition);
    }
}
exports.JoinConditionBuilder = JoinConditionBuilder;
//# sourceMappingURL=JoinMixin.js.map