Object.defineProperty(exports, "__esModule", { value: true });
exports.select = exports.from = exports.SelectQuery = void 0;
const Expressions_1 = require("../Expressions");
const FromFactor_1 = require("../FromFactor");
const RetrievalQuery_1 = require("./RetrievalQuery");
const Ordering_1 = require("../Ordering");
const Common_1 = require("./Common");
const JoinMixin_1 = require("./JoinMixin");
const Helpers_1 = require("../../Helpers");
class SelectQuery extends RetrievalQuery_1.RetrievalQuery {
    constructor() {
        super(...arguments);
        this._orderBys = [];
        this._groupBys = [];
        // #endregion
        // #region call-macro joinMixin()
        this._from = undefined;
        // #endregion
    }
    getState() {
        return Object.assign({
            orderBys: this._orderBys,
            whereCondition: this._whereCondition,
            havingCondition: this._havingCondition,
            groupBys: this._groupBys,
            from: this._from
        }, super.getState());
    }
    /**
     * Selects from a table. If previous tables are already specified, they are cross joined.
     * @param table The table to select from.
     */
    from(table) {
        this._from = FromFactor_1.FromFactor.crossJoin(this._from, table);
        this.lastFromItem = table;
        return this;
    }
    select(...args) {
        (0, Common_1.handleSelect)(this.lastFromItem, args, this.returningColumns, this.selectedExpressions);
        return this;
    }
    orderBy(...expressions) {
        const expressions2 = (() => {
            const firstExpr = expressions[0];
            if (firstExpr && (typeof firstExpr) === "function") {
                const expr = firstExpr(this.returningColumns);
                if (Array.isArray(expr))
                    return expr;
                return [expr];
            }
            return expressions;
        })();
        const exprs = expressions2.map(e => {
            if (typeof e === "string")
                return (0, Common_1.resolveColumnReference)(this.lastFromItem, e).asc();
            if (e instanceof Expressions_1.Expression)
                return e.asc();
            if ((0, Ordering_1.isOrderingAsc)(e))
                return (0, Common_1.resolveColumnReference)(this.lastFromItem, e.asc).asc();
            if ((0, Ordering_1.isOrderingDesc)(e))
                return (0, Common_1.resolveColumnReference)(this.lastFromItem, e.desc).desc();
            throw new Error(`Unexpected value in orderBy: '${e}'.`);
        });
        this._orderBys.push(...exprs);
        return this;
    }
    having(...args) {
        let expression = undefined;
        if ((typeof args[0]) === "function") {
            const func = args[0];
            expression = func(this.returningColumns);
        }
        else
            expression = (0, Expressions_1.toCondition)(this.lastFromItem, args);
        this._havingCondition = (0, Expressions_1.and)(this._havingCondition, expression);
        return this;
    }
    groupBy(...expressions) {
        const exprs = (() => {
            const firstExpr = expressions[0];
            if ((typeof firstExpr) === "function") {
                let result = firstExpr(this.returningColumns);
                if (Array.isArray(result))
                    return result;
                return [result];
            }
            else
                return expressions.map(e => (0, Common_1.resolveColumnReference)(this.lastFromItem, e));
        })();
        this._groupBys.push(...exprs);
        return this;
    }
    where(...args) {
        const expression = (0, Expressions_1.toCondition)(this.lastFromItem, args);
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, expression);
        return this;
    }
    /**
    * Adds negated where conditions.
    */
    whereNot(condition, ...conditions) {
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, condition.not(), ...conditions.map(c => c.not()));
        return this;
    }
    /**
    * Performs a full join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * for (row j in joined): if (cur has no row that matches j)
    * 	yield row(null, j)
    * ```
    */
    fullJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorFullJoin, fromItem);
    }
    /**
    * Performs a left join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * ```
    */
    leftJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorLeftJoin, fromItem);
    }
    /**
    * Performs an inner join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * ```
    */
    innerJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorInnerJoin, fromItem);
    }
}
exports.SelectQuery = SelectQuery;
/**
 * Creates a SELECT query that selects from the given table.
 * @param table The table to select from.
 */
function from(table) {
    const result = new SelectQuery();
    return result.from(table);
}
exports.from = from;
exports.select = (0, Helpers_1.secondWithTypeOfFirst)(new SelectQuery().select, function (...args) {
    const result = new SelectQuery();
    return result.select.call(result, ...args);
});
//# sourceMappingURL=SelectQuery.js.map