Object.defineProperty(exports, "__esModule", { value: true });
exports.UpdateQuery = exports.update = void 0;
const FromFactor_1 = require("../FromFactor");
const Expressions_1 = require("../Expressions");
const Query_1 = require("./Query");
const Common_1 = require("./Common");
const JoinMixin_1 = require("./JoinMixin");
const Helpers_1 = require("../../Helpers");
function update(table) {
    return new UpdateQuery(table);
}
exports.update = update;
class UpdateQuery extends Query_1.Query {
    constructor(table) {
        super();
        this._updatedColumns = {};
        // #endregion
        // #region call-macro joinMixin()
        this._from = undefined;
        this.lastFromItem = table;
        this._table = table;
    }
    getState() {
        return Object.assign({
            whereCondition: this._whereCondition,
            updatedColumns: this._updatedColumns,
            table: this._table,
            from: this._from
        }, super.getState());
    }
    /**
     * Selects from a table. If previous tables are already specified, they are cross joined.
     * These tables can be used to for condition or value expressions.
     * @param table The table to select from.
     */
    from(table) {
        this._from = FromFactor_1.FromFactor.crossJoin(this._from, table);
        this.lastFromItem = table;
        return this;
    }
    set(columnOrObject, value) {
        if (typeof columnOrObject === "string") {
            const column = (0, FromFactor_1.getColumn)(this._table, columnOrObject);
            const expr = (0, Expressions_1.normalize)(column.type, value);
            this._updatedColumns[columnOrObject] = expr;
        }
        else {
            for (const [name, value] of (0, Helpers_1.objectEntries)(columnOrObject)) {
                const column = (0, FromFactor_1.getColumn)(this._table, name);
                const expr = (0, Expressions_1.normalize)(column.type, value);
                this._updatedColumns[name] = expr;
            }
        }
        return this;
    }
    returning(...args) {
        (0, Common_1.handleSelect)(this.lastFromItem, args, this.returningColumns, this.selectedExpressions);
        return this;
    }
    where(...args) {
        const expression = (0, Expressions_1.toCondition)(this.lastFromItem, args);
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, expression);
        return this;
    }
    /**
    * Adds negated where conditions.
    */
    whereNot(condition, ...conditions) {
        this._whereCondition = (0, Expressions_1.and)(this._whereCondition, condition.not(), ...conditions.map(c => c.not()));
        return this;
    }
    /**
    * Performs a full join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * for (row j in joined): if (cur has no row that matches j)
    * 	yield row(null, j)
    * ```
    */
    fullJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorFullJoin, fromItem);
    }
    /**
    * Performs a left join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * for (row r in cur): if (joined has no row that matches r)
    * 	yield row(r, null)
    * ```
    */
    leftJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorLeftJoin, fromItem);
    }
    /**
    * Performs an inner join on the current query (`cur`) and a specified table (`joined`).
    * These rows are returned:
    * ```
    * for (row r in cur): for (row j in joined that matches r)
    * 	yield row(r, j)
    * ```
    */
    innerJoin(fromItem) {
        return (0, JoinMixin_1.doJoin)(this, this._from, newFrom => this._from = newFrom, FromFactor_1.FromFactorInnerJoin, fromItem);
    }
}
exports.UpdateQuery = UpdateQuery;
//# sourceMappingURL=UpdateQuery.js.map