import { SqlStatement, IsolationLevel } from './AST/SqlStatement';
import { Query } from "./AST/Queries/Query";
import { GetOutType, MapOutType } from "./AST/Types";
import { Row } from "./AST/FromFactor";
export interface SimpleDbQueryService {
    exec<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow>[]>;
    exec<TRow extends Row>(statement: SqlStatement): Promise<void>;
}
export interface DbQueryService extends SimpleDbQueryService {
    getExclusiveQueryService<T>(scope: (service: SimpleDbQueryService) => Promise<T>): Promise<T>;
}
export declare class SqlError extends Error {
    readonly code: QueryErrorCode;
    readonly sqlStatement: SqlStatement;
    constructor(m: string, code: QueryErrorCode, sqlStatement: SqlStatement);
}
export declare enum QueryErrorCode {
    ExpectedAtLeastOneRow = 0,
    MoreThanOneRowReturned = 1
}
/**
 * Provides methods to execute queries.
 */
export declare class DbQueryInterface {
    protected readonly queryService: SimpleDbQueryService;
    constructor(queryService: SimpleDbQueryService);
    exec<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow>[]>;
    exec<TRow extends Row>(statement: SqlStatement): Promise<void>;
    /**
     * Executes a query which only selects a single column and returns a list of values.
     * @param query A query which only selects a single column.
     */
    values<TRow extends Row, TColumn extends keyof TRow>(query: Query<TRow, TColumn>): Promise<GetOutType<TRow[TColumn]>[]>;
    /**
     * Returns the first row or `undefined` if there is none.
     * @param query The query to execute.
     */
    firstOrUndefined<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow> | undefined>;
    /**
     * Returns the first row. Throws an `SqlError` if there is none.
     * @param query The query to execute.
     */
    first<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow>>;
    /**
     * Returns the first row or `undefined` if there is none. Throws an `SqlError` if multiple rows are returned.
     * @param query The query to execute.
     */
    singleOrUndefined<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow> | undefined>;
    /**
     * Returns the first row. Throws an `SqlError` if no or multiple rows are returned.
     * @param query The query to execute.
     */
    single<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow>>;
    firstOrUndefinedValue<TRow extends Row, TColumn extends keyof TRow>(query: Query<TRow, TColumn>): Promise<GetOutType<TRow[TColumn]> | undefined>;
    firstValue<TRow extends Row, TColumn extends keyof TRow>(query: Query<TRow, TColumn>): Promise<GetOutType<TRow[TColumn]>>;
    singleValue<TRow extends Row, TColumn extends keyof TRow>(query: Query<TRow, TColumn>): Promise<GetOutType<TRow[TColumn]>>;
    singleOrUndefinedValue<TRow extends Row, TColumn extends keyof TRow>(query: Query<TRow, TColumn>): Promise<GetOutType<TRow[TColumn]> | undefined>;
}
export declare class TransactingQueryService extends DbQueryInterface {
    private _isTransactionOpen;
    /**
     * Checks whether the transaction has been closed, i.e. whether it has been rolled back or committed.
     */
    get isTransactionOpen(): boolean;
    exec<TRow extends Row>(query: Query<TRow, any>): Promise<MapOutType<TRow>[]>;
    exec<TRow extends Row>(statement: SqlStatement): Promise<void>;
    /**
     * Rolls the transaction back and closes the transaction.
     */
    rollback(): Promise<void>;
    /**
     * Commits the changes and closes the transaction.
     */
    commit(): Promise<void>;
}
export declare class DbConnection extends DbQueryInterface {
    protected readonly queryService: DbQueryService;
    constructor(queryService: DbQueryService);
    /**
     * Runs the `scope` callback within a transaction and returns its result.
     * The transaction will be committed automatically if the transaction has not been closed after the promise returned by the `scope` callback resolves.
     * The transaction will be rolled back if the promise rejects and the transaction has not been closed.
     * @param scope A callback that gets a query interface which runs in a new transaction.
     */
    transaction<T>(scope: (queryInterface: TransactingQueryService) => Promise<T>): Promise<T>;
    transaction<T>(options: {
        isolationLevel?: IsolationLevel;
    }, scope: (queryInterface: TransactingQueryService) => Promise<T>): Promise<T>;
}
