Object.defineProperty(exports, "__esModule", { value: true });
exports.DbConnection = exports.TransactingQueryService = exports.DbQueryInterface = exports.QueryErrorCode = exports.SqlError = void 0;
const SqlStatement_1 = require("./AST/SqlStatement");
const RetrievalQuery_1 = require("./AST/Queries/RetrievalQuery");
class SqlError extends Error {
    constructor(m, code, sqlStatement) {
        super(m);
        this.code = code;
        this.sqlStatement = sqlStatement;
        Object.setPrototypeOf(this, SqlError.prototype);
    }
}
exports.SqlError = SqlError;
var QueryErrorCode;
(function (QueryErrorCode) {
    QueryErrorCode[QueryErrorCode["ExpectedAtLeastOneRow"] = 0] = "ExpectedAtLeastOneRow";
    QueryErrorCode[QueryErrorCode["MoreThanOneRowReturned"] = 1] = "MoreThanOneRowReturned";
})(QueryErrorCode = exports.QueryErrorCode || (exports.QueryErrorCode = {}));
function expectedAtLeastOneRow(sqlStatement) {
    return new SqlError("Expected at least one row, but none were returned.", QueryErrorCode.ExpectedAtLeastOneRow, sqlStatement);
}
function moreThanOneRowReturned(sqlStatement) {
    return new SqlError("More than one row returned.", QueryErrorCode.MoreThanOneRowReturned, sqlStatement);
}
/**
 * Provides methods to execute queries.
 */
class DbQueryInterface {
    constructor(queryService) {
        this.queryService = queryService;
    }
    exec(statement) {
        return this.queryService.exec(statement);
    }
    /**
     * Executes a query which only selects a single column and returns a list of values.
     * @param query A query which only selects a single column.
     */
    async values(query) {
        const rows = await this.exec(query);
        return rows.map(r => r[query.singleColumn]);
    }
    /**
     * Returns the first row or `undefined` if there is none.
     * @param query The query to execute.
     */
    async firstOrUndefined(query) {
        if (query instanceof RetrievalQuery_1.RetrievalQuery)
            query = query.limit(1);
        const rows = await this.exec(query);
        return rows[0];
    }
    /**
     * Returns the first row. Throws an `SqlError` if there is none.
     * @param query The query to execute.
     */
    async first(query) {
        const result = await this.firstOrUndefined(query);
        if (!result)
            throw expectedAtLeastOneRow(query);
        return result;
    }
    /**
     * Returns the first row or `undefined` if there is none. Throws an `SqlError` if multiple rows are returned.
     * @param query The query to execute.
     */
    async singleOrUndefined(query) {
        if (query instanceof RetrievalQuery_1.RetrievalQuery)
            query = query.limit(2);
        const rows = await this.exec(query);
        if (rows.length === 0)
            return undefined;
        if (rows.length >= 2)
            throw moreThanOneRowReturned(query);
        return rows[0];
    }
    /**
     * Returns the first row. Throws an `SqlError` if no or multiple rows are returned.
     * @param query The query to execute.
     */
    async single(query) {
        if (query instanceof RetrievalQuery_1.RetrievalQuery)
            query = query.limit(2);
        const rows = await this.exec(query);
        if (rows.length === 0)
            throw expectedAtLeastOneRow(query);
        if (rows.length >= 2)
            throw moreThanOneRowReturned(query);
        return rows[0];
    }
    async firstOrUndefinedValue(query) {
        const row = await this.firstOrUndefined(query);
        if (!row)
            return undefined;
        return row[query.singleColumn];
    }
    async firstValue(query) {
        const row = await this.first(query);
        return row[query.singleColumn];
    }
    async singleValue(query) {
        const row = await this.single(query);
        return row[query.singleColumn];
    }
    async singleOrUndefinedValue(query) {
        const row = await this.singleOrUndefined(query);
        if (!row)
            return undefined;
        return row[query.singleColumn];
    }
}
exports.DbQueryInterface = DbQueryInterface;
class TransactingQueryService extends DbQueryInterface {
    constructor() {
        super(...arguments);
        this._isTransactionOpen = true;
    }
    /**
     * Checks whether the transaction has been closed, i.e. whether it has been rolled back or committed.
     */
    get isTransactionOpen() { return this._isTransactionOpen; }
    exec(statement) {
        if (!this._isTransactionOpen)
            throw new Error(`Could not execute statement '${statement}': Transaction already has been either committed or rolled back.`);
        return super.exec(statement);
    }
    /**
     * Rolls the transaction back and closes the transaction.
     */
    async rollback() {
        await this.exec(new SqlStatement_1.RollbackTransactionStatement());
        this._isTransactionOpen = false;
    }
    /**
     * Commits the changes and closes the transaction.
     */
    async commit() {
        await this.exec(new SqlStatement_1.CommitTransactionStatement());
        this._isTransactionOpen = false;
    }
}
exports.TransactingQueryService = TransactingQueryService;
class DbConnection extends DbQueryInterface {
    constructor(queryService) {
        super(queryService);
        this.queryService = queryService;
    }
    async transaction(scopeOrOptions, scope) {
        const actualScope = (typeof scopeOrOptions === "object") ? scope : scopeOrOptions;
        const options = (typeof scopeOrOptions === "object") ? scopeOrOptions : { isolationLevel: undefined };
        if (actualScope === undefined)
            throw new Error("No scope given.");
        return this.queryService.getExclusiveQueryService(async (s) => {
            const queryInterface = new TransactingQueryService(s);
            await queryInterface.exec(new SqlStatement_1.StartTransactionStatement(options.isolationLevel));
            let error = undefined;
            let result;
            try {
                result = { success: true, result: await actualScope(queryInterface) };
            }
            catch (err) {
                result = { success: false, error: err };
            }
            if (result.success) {
                if (queryInterface.isTransactionOpen)
                    await queryInterface.commit();
                return result.result;
            }
            else {
                if (queryInterface.isTransactionOpen)
                    await queryInterface.rollback();
                throw result.error;
            }
        });
    }
}
exports.DbConnection = DbConnection;
//# sourceMappingURL=DbConnection.js.map