Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlGenerator = void 0;
const RetrievalQuery_1 = require("./AST/Queries/RetrievalQuery");
const ValuesQuery_1 = require("./AST/Queries/ValuesQuery");
const InsertQuery_1 = require("./AST/Queries/InsertQuery");
const UpdateQuery_1 = require("./AST/Queries/UpdateQuery");
const DeleteQuery_1 = require("./AST/Queries/DeleteQuery");
const SqlStatement_1 = require("./AST/SqlStatement");
const FromFactor_1 = require("./AST/FromFactor");
const SelectQuery_1 = require("./AST/Queries/SelectQuery");
const Exprs = require("./AST/Expressions");
const Table_1 = require("./AST/Table");
const Ordering_1 = require("./AST/Ordering");
const Helpers_1 = require("./Helpers");
// IMPORTANT IMPLEMENTATION NOTICE: expressions must be generated in the order they appear in the final sql string!
// Otherwise prepared statements with parameters get messed up.
class SqlGenerator {
    constructor(options = {}) {
        this.options = options;
        this.fromFactorDispatcherInitialized = false;
        this.fromFactorDispatcher = new Helpers_1.DynamicDispatcher();
        this.expressionDispatcherInitialized = false;
        this.expressionDispatcher = new Helpers_1.DynamicDispatcher();
    }
    toSql(statement) {
        const context = { parameters: [] };
        const sql = this.transformToSql(statement, context);
        return { sql, parameters: context.parameters };
    }
    transformToSql(statement, context) {
        if (statement instanceof SelectQuery_1.SelectQuery)
            return this.transformSelectQueryToSql(statement, context);
        if (statement instanceof UpdateQuery_1.UpdateQuery)
            return this.transformUpdateQueryToSql(statement, context);
        if (statement instanceof InsertQuery_1.InsertQuery)
            return this.transformInsertQueryToSql(statement, context);
        if (statement instanceof DeleteQuery_1.DeleteQuery)
            return this.transformDeleteQueryToSql(statement, context);
        if (statement instanceof ValuesQuery_1.ValuesQuery)
            return this.transformValuesQueryToSql(statement, context);
        if (statement instanceof RetrievalQuery_1.UnionQuery)
            return this.transformUnionQueyToSql(statement, context);
        if (statement instanceof SqlStatement_1.StartTransactionStatement)
            return "BEGIN";
        if (statement instanceof SqlStatement_1.CommitTransactionStatement)
            return "COMMIT";
        if (statement instanceof SqlStatement_1.RollbackTransactionStatement)
            return "ROLLBACK";
        throw new Error(`Unsupported query: ${statement}`);
    }
    createExpressionContext(froms, context) {
        if (!this.options.shortenColumnNameIfUnambigous)
            return { isColumnNameUnambigous: name => false, resolveNamedExpression: false, context };
        let allColumns = [];
        for (const from of froms) {
            if (!from)
                continue;
            const allFromFactors = FromFactor_1.FromFactor.getAllFromFactors(from);
            allColumns = allColumns.concat(...allFromFactors.map(f => (0, Helpers_1.objectValues)(f.$columns)));
        }
        const set = new Set();
        const duplicates = new Set();
        for (const col of allColumns) {
            if (set.has(col.name))
                duplicates.add(col.name);
            else
                set.add(col.name);
        }
        return { isColumnNameUnambigous: name => !duplicates.has(name), resolveNamedExpression: false, context };
    }
    transformUnionQueyToSql(query, context) {
        const query1Sql = this.transformToSql(query.query1, context);
        const query2Sql = this.transformToSql(query.query2, context);
        return `(${query1Sql}) UNION ${query.isUnionAll ? "ALL " : ""}(${query2Sql})`;
    }
    transformValuesQueryToSql(query, context) {
        const columns = query.columns;
        if (query.values.length === 0) {
            return `SELECT ${Object.keys(columns).map(k => "null").join(", ")} WHERE false`;
        }
        let sql = "VALUES ";
        const expressionContext = { isColumnNameUnambigous: c => true, resolveNamedExpression: true, context };
        sql += query.values.map(v => `(${(0, Helpers_1.objectEntries)(columns).map(([name, type]) => this.escapeValue(new Exprs.ValueExpression(type, v[name]), expressionContext)).join(", ")})`).join(", ");
        return sql;
    }
    transformInsertQueryToSql(query, context) {
        const data = query.getState();
        let sql = `INSERT INTO ${this.referToFromItem(data.table)}`;
        const expressionContext = { isColumnNameUnambigous: c => true, resolveNamedExpression: true, context };
        if (Array.isArray(data.values)) {
            if (data.values.length === 0) {
                sql += ` (SELECT null WHERE false)`;
            }
            else {
                const firstRow = data.values[0];
                const columnNames = Object.keys(firstRow);
                for (const selectedCol of columnNames) {
                    if (!(selectedCol in data.table.$columns))
                        throw new Error(`Column '${selectedCol}' does not exist on table '${data.table}'.`);
                }
                sql += `(${columnNames.map(c => this.quoteColumnName(c)).join(", ")}) VALUES `;
                sql += data.values.map(v => `(${columnNames.map(colName => this.escapeValue(Exprs.val(v[colName], data.table.$columns[colName].type), expressionContext)).join(", ")})`).join(", ");
            }
        }
        else {
            const s = data.values.getState();
            const columnNames = s.selected.map(selectedCol => {
                if (selectedCol instanceof Exprs.AllExpression)
                    throw new Error("AllExpressions in an insert into query are not supported.");
                if (!(selectedCol.name in data.table.$columns))
                    throw new Error(`Column '${selectedCol.name}' does not exist on table '${data.table}'.`);
                return selectedCol.name;
            });
            sql += `(${columnNames.map(c => this.quoteColumnName(c)).join(", ")}) ${this.transformToSql(data.values, context)}`;
        }
        if (data.selected.length > 0) {
            sql += " RETURNING " + this.toSelectStatementStr(data.selected, expressionContext);
        }
        return sql;
    }
    transformDeleteQueryToSql(query, context) {
        const data = query.getState();
        let sql = `DELETE FROM ${this.referToFromItem(data.table)}`;
        const expressionContext = this.createExpressionContext([data.using, data.table], context);
        if (data.using) {
            sql += ` USING ${this.fromToSql(data.using, expressionContext)}`;
        }
        if (!data.whereCondition)
            throw new Error("Delete Queries must have a where condition.");
        sql += ` WHERE ${this.expressionToSql(data.whereCondition, expressionContext)}`;
        if (data.selected.length > 0) {
            sql += " RETURNING " + this.toSelectStatementStr(data.selected, expressionContext);
        }
        return sql;
    }
    transformUpdateQueryToSql(query, context) {
        const data = query.getState();
        let sql = `UPDATE ${this.referToFromItem(data.table)}`;
        const expressionContext = this.createExpressionContext([data.from, data.table], context);
        sql += " SET " + (0, Helpers_1.objectEntries)(data.updatedColumns).map(([name, value]) => `${this.quoteColumnName(name)} = ${this.expressionToSql(value, expressionContext)}`).join(", ");
        if (data.from) {
            sql += ` FROM ${this.fromToSql(data.from, expressionContext)}`;
        }
        if (!data.whereCondition)
            throw new Error("Update Queries must have a where condition.");
        sql += ` WHERE ${this.expressionToSql(data.whereCondition, expressionContext)}`;
        if (data.selected.length > 0) {
            sql += " RETURNING " + this.toSelectStatementStr(data.selected, expressionContext);
        }
        return sql;
    }
    transformSelectQueryToSql(query, context) {
        const data = query.getState();
        let sql = "SELECT";
        const expressionContext = this.createExpressionContext([data.from], context);
        if (data.selected.length > 0) {
            sql += " " + this.toSelectStatementStr(data.selected, expressionContext);
        }
        if (data.from) {
            sql += ` FROM ${this.fromToSql(data.from, expressionContext)}`;
        }
        if (data.whereCondition) {
            sql += ` WHERE ${this.expressionToSql(data.whereCondition, expressionContext)}`;
        }
        if (data.groupBys.length > 0) {
            const exprCt = { isColumnNameUnambigous: expressionContext.isColumnNameUnambigous, resolveNamedExpression: false, context: context };
            const statements = data.groupBys.map(e => this.expressionToSql(e, exprCt)).join(", ");
            sql += ` GROUP BY ${statements}`;
        }
        if (data.havingCondition) {
            const havingCondition = this.expressionToSql(data.havingCondition, { isColumnNameUnambigous: expressionContext.isColumnNameUnambigous, resolveNamedExpression: true, context: context });
            sql += ` HAVING ${havingCondition}`;
        }
        if (data.orderBys.length > 0) {
            sql += " ORDER BY " + data.orderBys.map(d => (0, Ordering_1.isOrderingAsc)(d) ? this.expressionToSql(d.asc, expressionContext) : (this.expressionToSql(d.desc, expressionContext) + " DESC"));
        }
        if (data.limit) {
            sql += " LIMIT " + this.expressionToSql(data.limit, expressionContext);
        }
        if (data.offset) {
            sql += " OFFSET " + this.expressionToSql(data.offset, expressionContext);
        }
        return sql;
    }
    toSelectStatementStr(selected, context) {
        return selected.map(expr => {
            if (expr instanceof Exprs.NamedExpressionWrapper) {
                return `${this.expressionToSql(expr.expression, context)} AS ${this.quoteColumnName(expr.name)}`;
            }
            return this.expressionToSql(expr, context);
        }).join(", ");
    }
    quoteSchemaName(name) {
        return this.quoteSchemaOrTableOrColumnName(name);
    }
    quoteTableName(name) {
        return this.quoteSchemaOrTableOrColumnName(name);
    }
    quoteColumnName(name) {
        return this.quoteSchemaOrTableOrColumnName(name);
    }
    referToFromItem(fromItem, includeSchema = true) {
        if (fromItem instanceof Table_1.Table) {
            let result = "";
            if (fromItem.$name.schema && includeSchema)
                result += this.quoteSchemaName(fromItem.$name.schema) + ".";
            result += this.quoteTableName(fromItem.$name.name);
            return result;
        }
        else if (fromItem instanceof FromFactor_1.NamedFromItem || fromItem instanceof FromFactor_1.QueryFromItem) {
            return this.quoteTableName(fromItem.$name);
        }
        throw "Unsupported";
    }
    fromToSql(f, context) {
        if (!this.fromFactorDispatcherInitialized) {
            this.fromFactorDispatcherInitialized = true;
            this.fromFactorDispatcher
                .register(Table_1.Table, (f, context) => {
                return this.referToFromItem(f);
            })
                .register(FromFactor_1.NamedFromItem, (f, context) => {
                const oldName = this.referToFromItem(f.fromItem);
                const newName = this.referToFromItem(f);
                let sql = `${oldName} AS ${newName}`;
                if ((0, FromFactor_1.isCastToColumns)(f))
                    sql += `(${Object.keys(f.$columns).map(n => this.quoteColumnName(n)).join(", ")})`;
                return sql;
            })
                .register(FromFactor_1.QueryFromItem, (f, context) => {
                const newName = this.referToFromItem(f);
                let sql = `(${this.transformToSql(f.query, context.context)}) AS ${newName}`;
                if ((0, FromFactor_1.isCastToColumns)(f))
                    sql += `(${Object.keys(f.$columns).map(n => this.quoteColumnName(n)).join(", ")})`;
                return sql;
            })
                .register(FromFactor_1.FromFactorAbstractJoin, (f, context) => {
                const left = this.fromToSql(f.leftArg, context);
                const right = this.fromToSql(f.rightArg, context);
                let sql;
                if (f instanceof FromFactor_1.FromFactorAbstractConditionalJoin) {
                    const condition = this.expressionToSql(f.joinCondition, context);
                    let join;
                    if (f instanceof FromFactor_1.FromFactorFullJoin)
                        join = "FULL JOIN";
                    else if (f instanceof FromFactor_1.FromFactorInnerJoin)
                        join = "JOIN";
                    else if (f instanceof FromFactor_1.FromFactorLeftJoin)
                        join = "LEFT JOIN";
                    else
                        throw new Error("Unsupported join");
                    return `${left} ${join} ${right} ON ${condition}`;
                }
                else if (f instanceof FromFactor_1.FromFactorCrossJoin) {
                    return `${left} CROSS JOIN ${right}`;
                }
                else
                    throw new Error("Unknown join.");
            });
        }
        return this.fromFactorDispatcher.dispatch(f, context);
    }
    escapeValue(expr, context) {
        const val = expr.value;
        const serialized = expr.type.serialize(val);
        context.context.parameters.push(serialized);
        return "?";
    }
    expressionToSql(e, context, parentPrecedenceLevel = 1000) {
        const result = this.expressionToSqlAutoParenthesis(e, context);
        if (e.precedenceLevel > parentPrecedenceLevel)
            return `(${result})`;
        return result;
    }
    expressionToSqlAutoParenthesis(e, context) {
        if (!this.expressionDispatcherInitialized) {
            this.expressionDispatcherInitialized = true;
            this.expressionDispatcher
                .register(Exprs.NamedExpressionWrapper, (e, context) => {
                if (context.resolveNamedExpression)
                    return this.expressionToSql(e.expression, context);
                else
                    return this.quoteColumnName(e.name);
            })
                .register(Exprs.ValueExpression, (e, context) => this.escapeValue(e, context))
                .register(Exprs.Column, (e, context) => {
                const columnName = this.quoteColumnName(e.name);
                if (context.isColumnNameUnambigous(e.name))
                    return columnName;
                const tableName = this.referToFromItem(e.fromItem);
                return `${tableName}.${columnName}`;
            })
                .register(Exprs.FromItemExpression, (e, context) => {
                const tableName = this.referToFromItem(e.fromItem, false);
                return tableName;
            })
                .register(Exprs.AllExpression, (e, context) => {
                const tableName = this.referToFromItem(e.fromItem);
                return `${tableName}.*`;
            })
                .register(Exprs.BinaryOperatorExpression, (e, context) => {
                const operator = e.operator;
                const left = this.expressionToSql(e.left, context, e.precedenceLevel);
                const right = this.expressionToSql(e.right, context, e.precedenceLevel);
                return `${left} ${operator} ${right}`;
            })
                .register(Exprs.KnownFunctionInvocation, (e, context) => {
                const functionName = e.functionName.toUpperCase();
                const args = e.args.map(arg => this.expressionToSql(arg, context)).join(", ");
                return `${functionName}(${args})`;
            })
                .register(Exprs.RetrievalQueryAsExpression, (e, context) => `(${this.transformToSql(e.query, context.context)})`)
                .register(Exprs.IsInValuesExpression, (e, context) => {
                if (e.values.length === 0)
                    return this.expressionToSql(Exprs.val(false, true), context, e.precedenceLevel);
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `${arg} IN (${e.values.map(v => this.expressionToSql(v, context)).join(", ")})`;
            })
                .register(Exprs.IsInQueryExpression, (e, context) => {
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `${arg} IN (${this.transformToSql(e.query, context.context)})`;
            })
                .register(Exprs.LikeExpression, (e, context) => {
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `${arg} LIKE ${this.expressionToSql(e.like, context, e.precedenceLevel)}`;
            })
                .register(Exprs.NotExpression, (e, context) => {
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `NOT ${arg}`;
            })
                .register(Exprs.IsNullExpression, (e, context) => {
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `${arg} IS NULL`;
            })
                .register(Exprs.IsNotNullExpression, (e, context) => {
                const arg = this.expressionToSql(e.argument, context, e.precedenceLevel);
                return `${arg} IS NOT NULL`;
            })
                .register(Exprs.JsonPropertyAccess, (e, context) => {
                const arg = this.expressionToSql(e.expression, context, e.precedenceLevel);
                return `${arg}->${this.escapeValue(Exprs.val(e.key, true), context)}`;
            })
                .register(Exprs.CastExpression, (e, context) => {
                if (e.isHiddenCast)
                    return this.expressionToSql(e.expression, context, e.precedenceLevel);
                return this.expressionToSql(e.expression, context) + "::" + e.type.name;
            });
        }
        return this.expressionDispatcher.dispatch(e, context);
    }
}
exports.SqlGenerator = SqlGenerator;
//# sourceMappingURL=SqlGenerator.js.map